package io.p2vman.pilers;

import net.kyori.adventure.text.Component;
import org.bukkit.*;
import org.bukkit.block.Block;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitRunnable;

import java.util.*;
import java.util.function.Consumer;

public class Builder {
    private static final double DESIRED_DISTANCE = 20.0;
    private static final double RADIUS = (DESIRED_DISTANCE * 42) / (2 * Math.PI);
    private static final double CHUNK_DIAGONAL_HALF = Math.sqrt(16 * 16 + 16 * 16) / 2.0;
    public void build(int online, Consumer<Set<Long>> onComplete, Consumer<Component> chat) {
        World world = Bukkit.getWorlds().get(0);
        double d = DESIRED_DISTANCE;

        if (online < 6) {
            d = 30d;
        } else if (online == 6) {
            d = 25d;
        }

        double radius = (d * Math.min(online, 42)) / (2 * Math.PI);

        Set<Long> chunks = new HashSet<>();
        Set<Long> pillars = new HashSet<>();

        {
            int minChunkX = (int) Math.floor(-RADIUS / 16.0) - 1;
            int maxChunkX = (int) Math.floor(RADIUS / 16.0) + 1;
            int minChunkZ = (int) Math.floor(-RADIUS / 16.0) - 1;
            int maxChunkZ = (int) Math.floor(RADIUS / 16.0) + 1;

            double nearestX, nearestZ, centerX, centerZ;
            int chunkX, chunkZ;
            for (chunkX = minChunkX; chunkX <= maxChunkX; chunkX++) {
                for (chunkZ = minChunkZ; chunkZ <= maxChunkZ; chunkZ++) {
                    nearestX = clamp(0, chunkX * 16, chunkX * 16 + 15);
                    nearestZ = clamp(0, chunkZ * 16, chunkZ * 16 + 15);


                    if (nearestX * nearestX + nearestZ * nearestZ <= RADIUS * RADIUS) {
                        chunks.add(((long) chunkX << 32) | (chunkZ & 0xFFFFFFFFL));
                    }

                    centerX = chunkX * 16 + 8;
                    centerZ = chunkZ * 16 + 8;

                    if (Math.sqrt(centerX * centerX + centerZ * centerZ) + CHUNK_DIAGONAL_HALF <= RADIUS) {
                        chunks.add(((long) chunkX << 32) | (chunkZ & 0xFFFFFFFFL));
                    }
                }
            }
        }

        {
            double angle;
            int x, z;
            for (int i = 0; i < Math.min(online, 64); i++) {
                angle = 2 * Math.PI * i / Math.min(online, 64);
                x = (int) Math.round(radius * Math.cos(angle));
                z = (int) Math.round(radius * Math.sin(angle));

                pillars.add(((long) x << 32) | (z & 0xFFFFFFFFL));
            }
        }

        {
            Runnable onComp = () -> {
                Queue<Long> queue = new ArrayDeque<>(pillars);
                new BukkitRunnable() {
                    @Override
                    public void run() {
                        for (int i = 0; i < 40; i++) {
                            if (queue.isEmpty()) {
                                cancel();
                                if (onComplete != null) {
                                    onComplete.accept(pillars);
                                }
                                return;
                            }

                            Long entry = queue.poll();
                            if (entry == null) continue;

                            int x = (int) (entry >> 32);
                            int z = (int) (entry & 0xFFFFFFFFL);

                            for (int y = -10; y <  10; y++) {
                                world.getBlockAt(x, y, z).setType(Material.BEDROCK);
                            }

                            chat.accept(
                                    Component.text("Processed piler ")
                                            .append(Component.text(x))
                                            .append(Component.text(" "))
                                            .append(Component.text(z))
                            );
                        }
                    }
                }.runTaskTimer(JavaPlugin.getProvidingPlugin(Pilers.class), 0L, 1L);
            };


            Queue<Long> queue = new ArrayDeque<>(chunks);
            new BukkitRunnable() {
                @Override
                public void run() {
                    for (int i = 0; i < 10; i++) {
                        if (queue.isEmpty()) {
                            cancel();
                            onComp.run();
                            return;
                        }

                        Long entry = queue.poll();
                        if (entry == null) continue;

                        int chunkX = (int) (entry >> 32);
                        int chunkZ = (int) (entry & 0xFFFFFFFFL);


                        Chunk chunk = world.getChunkAt(chunkX, chunkZ);

                        for (int x = 0; x < 16; x++) {
                            for (int y = world.getMinHeight(); y < world.getMaxHeight(); y++) {
                                for (int z = 0; z < 16; z++) {
                                    Block block = chunk.getBlock(x, y, z);
                                    if (block.getType() != Material.AIR) {
                                        block.setType(Material.AIR, false);
                                    }
                                }
                            }
                        }

                        chat.accept(
                                Component.text("Processed chunk ")
                                        .append(Component.text(chunkX))
                                        .append(Component.text(" "))
                                        .append(Component.text(chunkZ))
                        );
                    }
                }
            }.runTaskTimer(JavaPlugin.getProvidingPlugin(Pilers.class), 0L, 1L);
        }
    }

    private static double clamp(double value, double min, double max) {
        if (value < min) return min;
        if (value > max) return max;
        return value;
    }
}
